/*
 *      Copyright Richard Gorton, 2007-2011 All rights reserved
 *
 *      Permission is granted to use, duplicate, and
 *      publish this text and program as long as it
 *      includes this entire comment block and limited
 *      rights reference.
 */
/*
 * Utility to extract paths and strata from the stored db's and analyze them,
 * looking for various characteristics.  Namely, various kinds of isomers.
 * Report various dynamic execution count statistics re: paths/strata
 * including those for the various isomers.
 * Future enhancement: look for subset paths
 *
 * For two paths (strata) to be isomers, they are vectors of the same length,
 * all of whose elements exist in both paths/strata.  That is, the sequences
 * ABCD and ABDC are isomers, while ABCD and ABCF are not.
 *
 * Kinds of Isomers to detect:
 * [Plain old] Isomer - no special characteristics
 * Rotational - If PathA is ABCD and PathB is BCDA, then they are rotational.
 *      Meaningless for length2
 * Palindromes - where PathA is a palindrome of PathB: ABCD <---> DCBA
 *      Meaningless for length2
 *
 */

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <zlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>

#include <string>

extern "C" {
#include "puredb/puredb_read.h"
}
#include "pathprof.h"
#include "servercomm.h"

#define ANALYZE_PATHS_MODE      0x1
#define ANALYZE_STRATA_MODE     0x2
#define VERBOSE_MODE            0x4
int     g_mode = 0;

#define PATHLOG_EXT     ".pathlog.txt"
#define STRATALOG_EXT   ".stratalog.txt"
#define PATHDB_EXT      ".pathinfo_db.p_dbp"
#define STRATADB_EXT    ".stratainfo_db.s_dbp"
#define STRATACONTENT_DATA      ".rawstrata.data"
#define STRATACONTENT_INDEX     ".rawstrata.idx"

std::string g_baseFile;
std::string *g_pathLog = NULL;
std::string *g_strataLog = NULL;
std::string *g_pathDB = NULL;
std::string *g_strataDB = NULL;
std::string *g_rawstrata_data = NULL;
std::string *g_rawstrata_index = NULL;

// palindromic isomers DO actually occur!
// Quite the surprise, indeed
typedef struct ISOMER_STATS {
    int         numberPaths;    // number of unique paths at this length
    int         numIsomers;     // unique isomers of said paths
    int         numRotationals; // how many are rotational isomes
    int         numPalindromes; // Possible, but seems highly unlikely
            // dynamic counts at this path/strata length
    unsigned long   dynTotal;   // all paths
    unsigned long   dynIsomers;
    unsigned long   dynRotational;
    unsigned long   dynPalindrome;
    } ISOMER_STATS;
struct PSTATS   g_pathStats;
struct ISOMER_STATS     g_uniquePaths[VLPATHLIM];
void        *g_rawPaths[VLPATHLIM];
void        *g_canonPaths[VLPATHLIM];
int         g_sortVecLen;       // Needed to sort

typedef struct  PATHINFO {
    unsigned long   tripCount;
    bool            isIsomer;
    bool            isRotational;
    bool            isPalindrome;
    } PATHINFO;
struct PATHINFO *g_pathInfo = NULL;

static PureDB   pathDB_pureR;

void
Usage(char *argv0)
{
    fprintf (stderr, "Usage: %s [optional switches] <file basename>\n", argv0);
    fprintf (stderr, "\tA tool to find isometric paths and strata collected from a whole program profile\n");
    fprintf (stderr, "\twhere <file basename> is <executable name>.<pid> of the output from the wppp_3tier tool\n");
    fprintf (stderr, "\tFor example: if the output files are whet.5686.*, then the file basename is whet.5686\n");
    fprintf (stderr, "\t\tanalyze paths (default)\n");
    fprintf (stderr, "\t-s\tanalyze strata\n");

}

static uint64_t
ReadLongFromPathDB(char *keystr)
{
    const char  *func = "ReadLongFromPathDB";
    char        dbkey[2048];
    datum       Key;
    datum       Value;
    uint64_t    fetchVal;
    void        *dbReadp;
    int rc;
    uint64_t    retval = 0;
    char        uncompBuf[1024 * 1024];
    uint64_t    uncompLen = sizeof (uncompBuf);

    snprintf (dbkey, sizeof(dbkey), "%s", keystr);
    Key.dptr = dbkey;
    Key.dsize = strlen(dbkey);
    Value.dptr = (char *)&fetchVal;
    Value.dsize = sizeof(fetchVal);
    
        // pathDB assumed to be open
    rc = puredb_find (&pathDB_pureR,
                       Key.dptr, Key.dsize,
                       (off_t *)&Value.dptr, (size_t *)&Value.dsize);
    if (rc != 0) {
        fprintf (stderr, "Problem in [%s] for key [%s]\n", func, keystr);
    } else {
        dbReadp = puredb_read (&pathDB_pureR, (off_t) Value.dptr, Value.dsize);
        retval = *(uint64_t *)dbReadp;
        puredb_read_free (dbReadp);
    }

    return retval;
}

static bool
ReadStructFromPathDB(char *keystr, void *valBuf, uint64_t valBufSize)
{
    const char  *func = "ReadStructFromPathDB";
    bool        success = true;
    char        dbkey[2048];
    datum       Key;
    datum       Value;
    void        *dbReadp;
    int         rc;

    snprintf (dbkey, sizeof(dbkey), "%s", keystr);
    Key.dptr = dbkey;
    Key.dsize = strlen(dbkey);
    Value.dptr = (char *)valBuf;
    Value.dsize = valBufSize;
    
        // pathDB assumed to be open
    rc = puredb_find (&pathDB_pureR,
                       Key.dptr, Key.dsize,
                       (off_t *)&Value.dptr, (size_t *)&Value.dsize);
    if (rc != 0) {
        success = false;
        fprintf (stderr, "Problem in [%s] for key [%s]\n", func, keystr);
    } else {
        dbReadp = puredb_read (&pathDB_pureR, (off_t) Value.dptr, Value.dsize);
        memcpy (valBuf, dbReadp, valBufSize);
        // retval = *(uint64_t *)dbReadp;
        puredb_read_free (dbReadp);
    }

    return success;
}

// Use g_sortVecLen to know how many addresses to compare
// Our layout is:
//      PathID (uint64)
//      Vector of g_sortVecLen Bb addresses (uint64)
static int
PathCmp (const void *item1, const void *item2)
{
    int         i;
    uint64_t    *p1;
    uint64_t    *p2;
    uint64_t    bb1;
    uint64_t    bb2;
    int         res;

    p1 = (uint64_t *)item1;
    p2 = (uint64_t *)item2;

    for (i = 0; i < g_sortVecLen; i++) {
        bb1 = p1[i + 1];
        bb2 = p2[i + 1];

        if (bb1 < bb2) {
            return -1;
        }
        if (bb2 < bb1) {
            return 1;
        }
    }

    return 0;
}

// For canonicalizing a Bb vector
static int
BbCmp (const void *item1, const void *item2)
{
    uint64_t    *p1;
    uint64_t    *p2;

    p1 = (uint64_t *) item1;
    p2 = (uint64_t *) item2;

    if (*p1 < *p2) {
        return -1;
    }

    if (*p1 > *p2) {
        return 1;
    }
        // This should never happen - it means a Bb vector had the
        // same address twice, which violates the definition of a path
    fprintf (stderr,
                "Multiple Bbs with same address.  Length: %d (<0x%lx>::<0x%lx>)\n",
                g_sortVecLen, *p1, *p2);
    // abort();
    return 0;
}

static int
arePalindromes(struct MAX_PATH_ELT *pathBlob_1, struct MAX_PATH_ELT *pathBlob_2, int pathLen)
{
    const char  *func = "areRotationals";
    int         i, j;
    int         retval = 0;

    j = pathLen - 1;
    for (i = 0; i < pathLen; i++) {
        if (pathBlob_1->dataBlob[i].BbAddress != pathBlob_2->dataBlob[j].BbAddress) {
            break;
        }
        j--;
    }

    if (i == pathLen) {
        retval = 1;
    }

done: ;
    return retval;
}

static int
areRotationals(struct MAX_PATH_ELT *pathBlob_1, struct MAX_PATH_ELT *pathBlob_2, int pathLen)
{
    const char  *func = "areRotationals";
    int         i, j;
    int         retval = 0;

        // Find the first point at which they match
    for (j = 0; j < pathLen; j++) {
        if (pathBlob_1->dataBlob[0].BbAddress == pathBlob_2->dataBlob[j].BbAddress) {
            break;
        }
    }

        // Sanity check
    if (pathBlob_1->dataBlob[0].BbAddress != pathBlob_2->dataBlob[j].BbAddress) {
        fprintf (stderr, "%s: Oops!\n", func);
        goto done;
    }

    for (i = 0; i < pathLen; i++) {
        if (pathBlob_1->dataBlob[i].BbAddress != pathBlob_2->dataBlob[j].BbAddress) {
            break;
        }
        j++;
        if (j >= pathLen) {
            j = 0;
        }
    }

    if (i == pathLen) {
        retval = 1;
    }
done: ;
    return retval;
}

// utility: dump time (seconds.microseconds)
// to facilitate some performance evaluation
void
reportTime(const char *notation)
{
    int rc;
    struct timeval      tv;

    gettimeofday(&tv, NULL);
    fprintf (stdout, "[%s] time is %ld.%ld\n", notation, tv.tv_sec, tv.tv_usec);
}

// utility: dump a specific path by id (not canonicalized)
// DB assumed to be closed at present
static void
FetchAndDumpRawPath(int pathId, int pathLen, struct MAX_PATH_ELT *maxPathBlob)
{
    char                dbkey[2048];
    int                 curPathSize;
    int                 i;
    int                 rc;

    rc = puredb_open (&pathDB_pureR, g_pathDB->c_str());
    if (rc != 0) {
        fprintf(stderr, "Weird failure opening [%s] - aborting\n", g_pathDB->c_str());
        exit(1);
    }
    memset (maxPathBlob, 0, sizeof(*maxPathBlob));
    snprintf (dbkey, sizeof(dbkey), "%s:%d", KEY_PATHID, pathId);
        // One BB_BLOB encapsulated in PATH_QUEUE_ELT
    curPathSize =  sizeof (struct PATH_QUEUE_ELT) + (sizeof (struct BB_BLOB) * (pathLen - 1));
    ReadStructFromPathDB(dbkey, maxPathBlob, curPathSize);

    if (maxPathBlob->pathID != pathId) {
        fprintf (stderr, "oops\n");
    }

    // Optionally dump the contents
    if ((g_mode & VERBOSE_MODE) == 0) {
        goto done;
    }
    fprintf (stdout, "\t%d", pathId);
    for (i = 0; i < pathLen; i++) {
        fprintf (stdout, "\t%p", maxPathBlob->dataBlob[i].BbAddress);
    }
    fprintf (stdout, "\n");

done: ;
    (void) puredb_close(&pathDB_pureR);
}

// utility: dump all paths of length pathLen
static void
dumpUniques(uint64_t **pList, int pathLen)
{
    int         i;
    uint64_t    *table;

    table = pList[pathLen];
    fprintf (stdout, "************* Dump of paths: length = %d *************\n", pathLen);
    fprintf (stdout, "PathID\tBb vector");
    for (i = 0; i < g_uniquePaths[pathLen].numberPaths; i++) {
        int     j;

        fprintf (stdout, "\n%ld", table[i*(pathLen + 1)]);
        for (j = 1; j <= pathLen; j++) {
            fprintf (stdout, "\t%p", (void *)table[i*(pathLen+1) + j]);
        }
    }
    fprintf (stdout, "\n");
}

// Sort and analyze the paths.
// This is the content that would seem to be amenable to OpenCL
// Raw content will be in both g_rawPaths and g_canonPaths
void
doPathAnalysis()
{
    const char  *func="doPathAnalysis";
    const char  *begin="beginning of doPathAnalysis";
    const char  *end="end of doPathAnalysis";
    uint64_t    *evalPathList;

    reportTime (begin);
        // iterate by path length
        // Paths of length 1 are not paricularly interesting, except for the
        // dynamic frequency counts.
        // For path lengths of 2, all isomers are both rotational and
        // palindromic isomers.
    for (int pathLen = 1; pathLen < VLPATHLIM; pathLen++) {
        int     numElts;
        int     sortEltSize;
        bool    addedDynValues;

        addedDynValues = false;
        evalPathList = (uint64_t *)g_canonPaths[pathLen];
        numElts = g_uniquePaths[pathLen].numberPaths;
        sortEltSize = sizeof(uint64_t) * (pathLen + 1);

        // Gaah - our lack of a specific length means we need to have
        // compare function use a global variable to know how many
        // Bb addresses to compare
        g_sortVecLen = pathLen;
        qsort((void *)evalPathList, numElts, sortEltSize, PathCmp);

        // fprintf (stdout, "\nIsomers of length %d\n", pathLen);
        // An isomer: two or identical canonicalized paths
        for (int i = 0; i < numElts; i++) {
            for (int j = i + 1; j < numElts; j++) {
                uint64_t        p1_idx;
                uint64_t        p2_idx;
                uint64_t        pathId_1;
                uint64_t        pathId_2;
                struct MAX_PATH_ELT pathBlob_1;
                struct MAX_PATH_ELT pathBlob_2;

                p1_idx = i*(pathLen + 1);
                p2_idx = j*(pathLen + 1);
                pathId_1 = evalPathList[p1_idx];
                pathId_2 = evalPathList[p2_idx];
                if (PathCmp(&evalPathList[p1_idx], &evalPathList[p2_idx]) == 0) {
                        // yes, it is an isomer
                    g_uniquePaths[pathLen].numIsomers++;
                    g_pathInfo[pathId_1].isIsomer = true;
                    g_pathInfo[pathId_2].isIsomer = true;

                        // Suck the full paths back out of the db
                    FetchAndDumpRawPath(pathId_1, pathLen, &pathBlob_1);
                    FetchAndDumpRawPath(pathId_2, pathLen, &pathBlob_2);
                    fprintf (stdout, "Isometric Paths: length: %d\tPathIds %ld::%ld\tCounts: %ld::%ld",
                                pathLen,
                                pathId_1,
                                pathId_2,
                                g_pathInfo[pathId_1].tripCount,
                                g_pathInfo[pathId_2].tripCount);

                        // Rotational isomers: ABCD and CDAB are rotational
                        // A special subcase
                    if (areRotationals(&pathBlob_1, &pathBlob_2, pathLen)) {
                        g_uniquePaths[pathLen].numRotationals++;
                        g_pathInfo[evalPathList[p1_idx]].isRotational = true;
                        g_pathInfo[evalPathList[p2_idx]].isRotational = true;
                        fprintf (stdout, "\tRotational");
                    }
                        // Palindrome isomers: ABCD and DCBA are rotational
                        // A special subcase.  Seems completely unlikely as this
                        // would require a number of computed jumps/calls in sequence
                        // A parser with a switch statement might do this
                    if (arePalindromes(&pathBlob_1, &pathBlob_2, pathLen)) {
                        g_uniquePaths[pathLen].numPalindromes++;
                        g_pathInfo[evalPathList[p1_idx]].isPalindrome = true;
                        g_pathInfo[evalPathList[p2_idx]].isPalindrome = true;
                        fprintf (stdout, "\tPalindromic");
                    }
                    fprintf (stdout, "\n");
                }       // are they isomers
            }   // iterate over j
#if 0
        // Full dump of whetstone shows paths 98, 102 (length 3) are isomers
        //      98      0x3a08109ce9    0x3a08109db8    0x3a0810a1d0
        //      102     0x3a08109ce9    0x3a08109db8    0x3a0810a1d0
        //
            fprintf (stdout, "%d", evalPathList[i*(pathLen + 1)]);
            for (j = 1; j <= pathLen; j++) {
                fprintf (stdout, "\t%p", evalPathList[i*(pathLen + 1) + j]);
            }
            fprintf (stdout, "\n");
            fflush (stdout);
#endif
        }       // iterate over i

        /*
         * Now that we have marked each path about whether or not they are isomers,
         * and isomer subtypes, collect the dynamic counts (by path length).
         * Do this as a separate pass in order to avoid both under and over-counting
         * of a particular path.
         */
        for (int i = 0; i < numElts; i++) {
            uint64_t    p1_idx;
            uint64_t    pathId;

            p1_idx = i*(pathLen + 1);
            pathId = evalPathList[p1_idx];
            g_uniquePaths[pathLen].dynTotal += g_pathInfo[pathId].tripCount;
            if (g_pathInfo[pathId].isIsomer) {
                g_uniquePaths[pathLen].dynIsomers += g_pathInfo[pathId].tripCount;
            }
            if (g_pathInfo[pathId].isRotational) {
                g_uniquePaths[pathLen].dynRotational += g_pathInfo[pathId].tripCount;
            }
            if (g_pathInfo[pathId].isPalindrome) {
                g_uniquePaths[pathLen].dynPalindrome += g_pathInfo[pathId].tripCount;
            }
        }   // each path at this length

    }   // iterate over length

    fprintf (stdout, "[isomer summary begin]\n");
    fprintf (stdout, "*** Isometric path summary (pairs of paths for isomer counts)***\n");
    fprintf (stdout, "\t\tDynamic\t\t\tDynamic\t\t\tDynamic\t\t\tDynamic\n");
    fprintf (stdout, "Path\tUnique\tpath\t\tAll\tIsomer\t\tRot.\tRot.\t\tPal.\tPal.\n");
    fprintf (stdout, "Length\tPaths\tCount\t\tIsomers\tCount\t\tIsomers\tCount\t\tIsomers\tCount\n");
    for (int i = 0; i < VLPATHLIM; i++) {
        if (g_uniquePaths[i].numberPaths != 0) {
            fprintf (stdout, "%d"           // path length
                             "\t%d\t%ld"    // num uniq paths, dynamic count of uniques
                             "\t\t%d\t%ld"  // isomers, dynamic count of all isomers
                             "\t\t%d\t%ld"  // rotational isomers, dynamic count of rotational
                             "\t\t%d\t%ld"  // palindrome isomers, dynamic count of palindrome
                             "\n",
                             i,
                             g_uniquePaths[i].numberPaths, g_uniquePaths[i].dynTotal,
                             g_uniquePaths[i].numIsomers, g_uniquePaths[i].dynIsomers,
                             g_uniquePaths[i].numRotationals, g_uniquePaths[i].dynRotational,
                             g_uniquePaths[i].numPalindromes, g_uniquePaths[i].dynPalindrome);
        }
    }
    fprintf (stdout, "[isomer summary end]\n");
    reportTime (end);
}

// Fill in g_uniquePaths table.
// BUGBUG FIXME TODO
// Ooops - at present, the path stats do not provide any information
// re: distribution of unique paths by length.  So this requires a
// full trip through the DB.  This ought to be done at collection time
// (simple enhancement)
// Save the tripcounts in g_tripCounts;
void
computeUniqueDistribution()
{
    const char  *func = "computeUniqueDistribution";
    const char  *begin = "beginning of computeUniqueDistribution";
    const char  *end = "end of computeUniqueDistribution";

    reportTime (begin);
    // add one: zero-based indices
    g_pathInfo = (struct PATHINFO *) calloc(g_pathStats.uniquePaths + 1, sizeof(struct PATHINFO));

    memset (g_uniquePaths, 0, sizeof (g_uniquePaths));
    for (int i = FIRST_PATH_NUM; i < g_pathStats.uniquePaths; i++) {
        char                    dbkey[2048];
        struct MAX_PATH_ELT     maxPathBlob;
        struct PATH_QUEUE_ELT   corePathInfo;
        int                     curPathSize;

        memset (&maxPathBlob, 0, sizeof(maxPathBlob));
        snprintf (dbkey, sizeof(dbkey), "%s:%d", KEY_PATHID, i);
        // One BB_BLOB encapsulated in PATH_QUEUE_ELT
        // Read it twice (ugh) once to get the path length, and then a "full" read
        // This to ensure we do not read off from unmapped content
        ReadStructFromPathDB(dbkey, &corePathInfo, sizeof(corePathInfo));
        curPathSize =  (corePathInfo.pathLen - 1) * (sizeof (struct BB_BLOB))
                        + sizeof(corePathInfo);

        ReadStructFromPathDB(dbkey, &maxPathBlob, curPathSize);

        g_uniquePaths[maxPathBlob.pathLen].numberPaths++;
        g_pathInfo[i].tripCount = maxPathBlob.tripCount;
    }
    reportTime (end);

}

void
setupPathAnalysis()
{
    const char  *func = "setupPathAnalysis";
    const char  *begin = "beginning of setupPathAnalysis";
    const char  *end = "end of setupPathAnalysis";
    int         rc;
    uint64_t    keyver;
    uint64_t    numpaths;
    bool        readOK;
    int         i;
    uint64_t    uniquePathLenMean;
    int         curPathSize;
    int         fillIndex[VLPATHLIM];

    reportTime (begin);
    rc = puredb_open (&pathDB_pureR, g_pathDB->c_str());
    if (rc != 0) {
        fprintf(stderr, "Weird failure opening [%s] - aborting\n", g_pathDB->c_str());
        exit(1);
    }

        // From paths.c:WritePathsDB()
        //      key             value
        //      KEY_VERNO       key version
        //      KEY_NUMPATHS    number paths
        //      KEY_PATHSTATS   path stat summary
        //      KEY_PATHID      path content

    keyver = ReadLongFromPathDB((char *)KEY_VERNO);
    if (keyver != KEY_VERSION) {
        fprintf(stderr, "Mismatch in db for key: [%s] got %ld\n", KEY_VERNO, keyver);
    }
    // fprintf(stdout, "key: [%s] got %d\n", KEY_VERNO, keyver);
    numpaths = ReadLongFromPathDB((char *)KEY_NUMPATHS);
    if (numpaths <= 100) {
        fprintf(stderr, "Probable error - only %ld paths - seems low\n", numpaths);
    }
    fprintf(stdout, "key: [%s] got %ld\n", KEY_NUMPATHS, numpaths);

        // Suck out path stat table
    readOK = ReadStructFromPathDB((char *)KEY_PATHSTATS, &g_pathStats, sizeof(g_pathStats));
    if (!readOK) {
        fprintf(stderr, "Error - could not read structure [%s]\n", KEY_PATHSTATS);
    }

        // Create the path tripcount 
        // BUGBUG FIXME TODO
        // Ooops - at present, the path stats do not provide any information
        // re: distribution of unique paths by length.  So this requires a
        // full trip through the DB.  This ought to be done at collection time
        // (simple enhancement)
    computeUniqueDistribution();


    fprintf (stdout, "\nPath definition: a path is defined as a sequence of basic blocks\n"
                "which do not repeat.  If the Bb sequence seen is AABCDCBADD, the paths are:\n"
                "\tA\tABCD\tCBAD\tD\n\n");

    fprintf (stdout, "Info: Number paths longer than %d Bbs: %ld\n", VLPATHLIM, g_pathStats.VLPaths);
    fprintf (stdout, "Info: Highest trip count going into ProcessPath: %ld\n", g_pathStats.highestTripCount);


    fprintf (stdout, "\nPath length distribution histogram\n");
    fprintf (stdout, "\t\tUnique\tDynamic\n");
    fprintf (stdout, "\tLength\tPaths\tCount\n");
    for (i = 0; i < VLPATHLIM; i++) {
        if (g_pathStats.pathDistribution[i] != 0) {
            fprintf (stdout, "\t%d\t%d\t%ld\n", i, g_uniquePaths[i].numberPaths,  g_pathStats.pathDistribution[i]);
        }
    }

    uniquePathLenMean = g_pathStats.uniquePathLenSum / g_pathStats.uniquePaths;
    fprintf (stdout, "\nSpreadsheet usable stats\n");
    fprintf (stdout, "sshdr\tNumber\tNumber\tNumber\tMax\tMax\tMean\tNumber\tMean\tStrata\n");
    fprintf (stdout,
        "sshdr\tInsns\tBbs\tPaths\tInsns\tBbs\tBb\tUnique\tBbs\tBuffers\n"
        "sshdr\t\t\t\tper\tper\tPath\tPaths\tUnique\t\n"
        "sshdr\t\t\t\tPath\tPath\tLen\t\tPath\t\n");

    fprintf (stdout, "ssdata\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\n",
        g_pathStats.totalInstructions, g_pathStats.totalBbs, g_pathStats.totalPaths,
        (uint64_t) g_pathStats.maxPathInsnLen, (uint64_t) g_pathStats.maxPathBbLen,
        g_pathStats.totalBbs / g_pathStats.totalPaths,
        g_pathStats.uniquePaths, uniquePathLenMean,
        g_pathStats.strataBuffers);

    memset (g_rawPaths, 0, sizeof(g_rawPaths));
    memset (g_canonPaths, 0, sizeof(g_canonPaths));
    memset (fillIndex, 0, sizeof(fillIndex));
        // Allocate the path structs to be sorted
    for (i = 0; i < VLPATHLIM; i++) {
        uint64_t        pathID;
        // basically sizeof(uint64_t) * (i+1) on Linux
        curPathSize = sizeof (pathID) + (sizeof (struct BB_BLOB) * i);
        g_rawPaths[i] = calloc (g_uniquePaths[i].numberPaths, curPathSize);
        g_canonPaths[i] = calloc (g_uniquePaths[i].numberPaths, curPathSize);
    }
        // Fill in the g_rawPaths table (reference)
        // And also the g_canonPaths table (which is not yet sorted)
        // Pull out the whole path structure, then copy relevant bits

    for (i = FIRST_PATH_NUM; i < g_pathStats.uniquePaths; i++) {
        char                    dbkey[2048];
        struct MAX_PATH_ELT     maxPathBlob;
        struct MAX_PATH_ELT     canonPathBlob;
        struct PATH_QUEUE_ELT   corePathInfo;
        int                     curPathSize;
        int                     currIndex;
        uint64_t                *rawArray;
        uint64_t                *canonArray;
        int                     pLen;
        int                     j;

        memset (&maxPathBlob, 0, sizeof(maxPathBlob));
        snprintf (dbkey, sizeof(dbkey), "%s:%d", KEY_PATHID, i);
        // One BB_BLOB encapsulated in PATH_QUEUE_ELT
        // Read it twice (ugh) once to get the path length, and then a "full" read
        // This to ensure we do not read off from unmapped content
        ReadStructFromPathDB(dbkey, &corePathInfo, sizeof(corePathInfo));
        curPathSize =  (corePathInfo.pathLen - 1) * (sizeof (struct BB_BLOB))
                        + sizeof(corePathInfo);

        ReadStructFromPathDB(dbkey, &maxPathBlob, curPathSize);

        pLen = maxPathBlob.pathLen;
        // canonicalize the Bb vector
        canonPathBlob = maxPathBlob;
        g_sortVecLen = pLen;
        qsort((void *)&canonPathBlob.dataBlob[0], pLen, sizeof(struct BB_BLOB), BbCmp);

        currIndex = fillIndex[pLen];
        rawArray = (uint64_t *) g_rawPaths[pLen];
        canonArray = (uint64_t *) g_canonPaths[pLen];
        if (rawArray == NULL) {
            continue;
        }

        // Path ID, then contents
        rawArray[currIndex] = maxPathBlob.pathID;
        canonArray[currIndex] = maxPathBlob.pathID;
        currIndex++;
        for (j = 0; j < pLen; j++) {
                // A bit slimy - know that void* is same size as uint64_t
            rawArray[currIndex] = (uint64_t) maxPathBlob.dataBlob[j].BbAddress;
            canonArray[currIndex] = (uint64_t) canonPathBlob.dataBlob[j].BbAddress;
            currIndex++;
        }


        fillIndex[pLen] = currIndex;
        // And a sanity check - 
        if (fillIndex[pLen] > g_pathStats.pathDistribution[pLen] * (pLen + 1)) {
            fprintf(stderr, "Something is screwy: plen: %d, pathid: %d\n", pLen, i);
        }
    }
    (void) puredb_close(&pathDB_pureR);
    reportTime (end);

}

void
doStrataAnalysis()
{
}


bool
fileExists(const char *pathName)
{
    struct stat stbuf;
    int         rc;
    bool        retval = false;

    rc = stat(pathName, &stbuf);
    if (rc == 0) {
        retval = true;
    }

    return retval;
}

bool
dataFilesOK(std::string &fname)
{
    bool        result = true;

    g_pathLog           = new std::string(fname + PATHLOG_EXT);
    g_strataLog         = new std::string(fname + STRATALOG_EXT);
    g_pathDB            = new std::string(fname + PATHDB_EXT);
    g_strataDB          = new std::string(fname + STRATADB_EXT);
    g_rawstrata_data    = new std::string(fname + STRATACONTENT_DATA);
    g_rawstrata_index   = new std::string(fname + STRATACONTENT_INDEX);

    if (!fileExists(g_pathLog->c_str())) {
        fprintf (stderr, "%s missing\n", g_pathLog->c_str());
        result = false;
    }
    if (!fileExists(g_strataLog->c_str())) {
        fprintf (stderr, "%s missing\n", g_strataLog->c_str());
        result = false;
    }
    if (!fileExists(g_pathDB->c_str())) {
        fprintf (stderr, "%s missing\n", g_pathDB->c_str());
        result = false;
    }
    if (!fileExists(g_strataDB->c_str())) {
        fprintf (stderr, "%s missing\n", g_strataDB->c_str());
        result = false;
    }
    if (!fileExists(g_rawstrata_data->c_str())) {
        fprintf (stderr, "%s missing\n", g_rawstrata_data->c_str());
        result = false;
    }
    if (!fileExists(g_rawstrata_index->c_str())) {
        fprintf (stderr, "%s missing\n", g_rawstrata_index->c_str());
        result = false;
    }

done:
    return result;
}

int
main(int argc, char* argv[])
{

    if (argc > 3) {
        Usage(argv[0]);
        exit(1);
    }

    g_mode = ANALYZE_PATHS_MODE;
    if (argc > 1) {
        if ((strcmp(argv[1], "-s") == 0)) {
            g_mode = ANALYZE_STRATA_MODE;
        } else if (strcmp(argv[1], "-v") == 0) {
            g_mode |= VERBOSE_MODE;
        }
    }

    g_baseFile = argv[argc - 1];
        // A complete set of files for a wpp (whole program profile) is:
        //      whet.5686.pathlog.txt
        //      whet.5686.stratalog.txt
        //      whet.5686.pathinfo_db.p_dbp
        //      whet.5686.rawstrata.data
        //      whet.5686.rawstrata.idx
        //      whet.5686.stratainfo_db.s_dbp
        //
    if (!dataFilesOK(g_baseFile)) {
        exit(1);
    }

        // Open files, look for isomers
    if ((g_mode & ANALYZE_PATHS_MODE) != ANALYZE_PATHS_MODE) {
        doStrataAnalysis();
    } else {
        setupPathAnalysis();
        doPathAnalysis();
    }

    return 0;
}

