/*
 *      Copyright, 2007-2011 (c)  Richard Gorton, All Rights Reserved.
 *
 *      Permission is granted to use, duplicate, and
 *      publish this text and program as long as it
 *      includes this entire comment block and limited
 *      rights reference.
 *
 *      Author: Rick Gorton     richard.gorton@amd.com
 *
 *      Layered strata (levels above simple paths) handling
 *
 *      Generates strata/count pairs and sends them to the strata compressor server.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <unistd.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>

#include "pathprof.h"
#include "servercomm.h"
#include "puredb/puredb_write.h"

extern char*    g_argV0;
extern uint64_t g_insnEpoch;
extern uint64_t g_pathEpoch;
extern struct SHARED_MEM*       g_compressSharedMem;

static void CheckHashFn();

static struct STRATUM_STATS     stratumStats;
static PureDBW  strataDB_pureW;
static int      finalFlush;
static FILE*    s_LogF;

#ifdef WPPP_DEBUG
static int      DBGCheckHash = 0;
#endif
#define HASH_CHECK_FREQ 1000
#define HASH_SORT_FREQ  1000

/*
 * Have the structures and variables mimic the path variables - it may be possible to
 * re-use and condense the code (recurse?) at some future point.
 *
 * However, for this portion of the code, we do not have a quick cache - it is
 * only meaningful in the context of a path profiler (and not a whole program path profiler)
 */
static union SPB_U              s_tmpStratum[VLPATHLIM + 1];
static struct STRATUM_ELT*      s_prevStratum = NULL;
static struct SHARED_MEM        s_strataBuf;
static uint64_t    s_rawIdx;
static unsigned short   s_addHash;
static int              s_curStrataLen;

/*
 * Use a large hash table, as in ProcessPath()
 */
typedef struct STRATALAYER_ENT {
    struct STRATUM_ELT*         item;
    struct STRATALAYER_ENT*     next;
    } STRATALAYER_ENT;
#define SLHSIZE         (256 * 1024)
static struct STRATALAYER_ENT** SLHash;

// Descending sort function
static int HChainCmp (const void *item1, const void* item2)
{
    struct STRATALAYER_ENT*     L1_pp;
    struct STRATALAYER_ENT*     L2_pp;
    struct STRATUM_ELT*         E1_p;
    struct STRATUM_ELT*         E2_p;
    int res;

    L1_pp = (struct STRATALAYER_ENT*) item1;
    L2_pp = (struct STRATALAYER_ENT*) item2;
    E1_p = (struct STRATUM_ELT*) L1_pp->item;
    E2_p = (struct STRATUM_ELT*) L2_pp->item;

    if (E2_p->tripCount > E1_p->tripCount)
        res = 1;
    else if (E2_p->tripCount < E1_p->tripCount)
        res = -1;
    else
        res = 0;

    return res;
}

#define DEPTH_TRIGGER   1
#define TRIPS_TRIGGER   100000
typedef struct SORTREC {
    struct STRATALAYER_ENT*     pse;
    uint64_t                        trips;
    } SORTREC;

static void RestructureHash()
{
    uint64_t        maxTrips;
    int         i;

    for (i = 0; i < SLHSIZE; i++) {
        struct STRATALAYER_ENT*     base;
        int     depth;

        depth = 0;
        maxTrips = 0;
        base = SLHash[i];
        while (base != NULL) {
            uint64_t        curTrips;

            curTrips = base->item->tripCount;
            if (curTrips > maxTrips) {
                maxTrips = curTrips;
            }
            depth++;
            base = base->next;
        }

        /*
         * Do not reorder for trivial chains.  Either too short a length, or all members
         * having trivial trip counts.
         * For performance reasons, use a custom sort instead of qsort.
         */
        if ((depth > DEPTH_TRIGGER) && (maxTrips > TRIPS_TRIGGER)) {
            struct SORTREC*     toBeSorted;
            struct SORTREC      newrec;
            struct STRATALAYER_ENT*     walker;
            int j;
            int k;
            int l;

                // Fill a linear table
            base = SLHash[i];
            toBeSorted = (struct SORTREC*) calloc (depth, sizeof(*toBeSorted));
            for (j = 0; j < depth; j++) {
                newrec.pse = base;
                newrec.trips = base->item->tripCount;
                base = base->next;
                // Simply tack trivial entries onto the end, and we are done
                if (newrec.trips < TRIPS_TRIGGER) {
                    toBeSorted[j] = newrec;
                    continue;
                }

                // Find correct spot to insert, and shift the rest
                k = 0;
                while ((k < j) && (toBeSorted[k].trips > newrec.trips)) {
                    k++;
                }

                // At the correct location: shift remainder
                l = j;
                while ((l > k) && (l > 0)) {
                    toBeSorted[l] = toBeSorted[l - 1];
                    l--;
                }

                // Insert the new
                toBeSorted[k] = newrec;
            }

                // then unpack the table back into the hash chain
            SLHash[i] = toBeSorted[0].pse;
            walker = toBeSorted[0].pse;
            for (j = 0; j < depth - 1; j++) {
                walker->next = toBeSorted[j + 1].pse;
                walker = walker->next;
            }
            walker->next = NULL;        // Force null termination

                // Free the table
            free (toBeSorted);
        }       // Was the hash slot worth sorting?
    }   // Each hash slot
}
// forcewrite is 0 normally, 1 for Fini() cleanup
#define WRITE_STRATA    0
#define FLUSH_STRATA    1
typedef struct IDTRIP {
    uint64_t       ident;
    uint64_t       trips;
    } IDTRIP;
static void WriteStrataToCompressionServer (uint64_t strataIdent, uint64_t layerTrips, int forcewrite)
{
    int         layerSize;
    int         rc = 0;
    int         sendqty = 0;
    struct IDTRIP       strataRecord;

    strataRecord.ident = strataIdent;
    strataRecord.trips = layerTrips;

    if ((s_rawIdx + sizeof(strataRecord) > DB_BUF_SIZE) || (forcewrite != 0)) {
        int     res;
        uint64_t   cBufSize;
        char    opFlags;

        opFlags = 0x0;
        if (stratumStats.numDBEntries == 0) {
            opFlags |= OP_OPEN_APPEND;
        }
        if (forcewrite != 0) {
            opFlags |= OP_FLUSH_CLOSE;
        }
        stratumStats.numDBEntries++;

        /*
         * Periodically check for a brain-dead hash function
         * And even if it is good, there are pathological cases (hmmer/sjeng)
         * where there seem to be significant phase changes during the program, whereby
         * a layer seen early is relegated to the end of the hash chain, and then
         * becomes hot again.  Fix this up somewhat with a periodic sort (by frequency)
         * of each hash slot, putting the hottest ones first.
         * Doing this is a significant performance divot remover.
         * What formerly took more than a week on an Athlon64 X2 5200+ (without completing)
         * is shortened significantly.
         * The current hash function seems to work pretty well on non-trivial apps.
         */

#ifdef WPPP_DEBUG
        if (DBGCheckHash != 0) {
            if (stratumStats.numDBEntries % HASH_CHECK_FREQ == 0) {
                CheckHashFn();
            }
        }
#endif

                // Re-order the hash slots.  Say every 1000 entries (about 1Gb of strata content)
        if (stratumStats.numDBEntries % HASH_SORT_FREQ == 0) {
            RestructureHash();
        }

        // Build up a message for the server to compress
        s_strataBuf.dataHdr.version = THIS_VERSION;
        s_strataBuf.dataHdr.operation = opFlags;
        s_strataBuf.dataHdr.flags = FLAG_COMPRESS_9;
        s_strataBuf.dataHdr.payloadSize = s_rawIdx;
        s_strataBuf.dataHdr.startInsnEpoch = g_insnEpoch;
        s_strataBuf.dataHdr.startPathEpoch = g_pathEpoch;
        s_strataBuf.dataHdr.startStrataEpoch = stratumStats.layer0cycles;

        // Send the message to the server, once server is available
        WaitUntilReadyToFill (g_compressSharedMem);
        memcpy ((void*) g_compressSharedMem, (void*) &s_strataBuf, sizeof(s_strataBuf));
        MarkDataFilled (g_compressSharedMem);

        // reset for further use
        s_rawIdx = 0;
    }

        // copy new info into buffer
    memcpy ((void*) &s_strataBuf.rawDBBuf[s_rawIdx], (void*) &strataRecord, sizeof(strataRecord));
    s_rawIdx += sizeof(strataRecord);
}

static inline struct STRATUM_ELT* AllocateElt (uint64_t sLen)
{
    int layerSize;
    struct STRATUM_ELT* newElt;

        // sLen is the current length of the path; subtract two
        // for the first two elements being in the structure already
    layerSize = sizeof(struct STRATUM_ELT) + ((sLen - 2) * sizeof(struct SMALL_PATH_BLOB));
    newElt = (struct STRATUM_ELT*) malloc (layerSize);

    return newElt;
}

/*
 * we don't really know how many unique strata layers there are.
 * for now guess that a good hash is
 *      ((addHash << 4) + stratumLen) & 0x3ffff
 * since Lengths are small; hashes can be up to 64K (but are typically less)
 * stratumID is 0 at this point...
 * 21 Aug: length is now bounded to be 2,3,4; don't use length as much
 * The addHash value is 16-bit truncated sum of path ids.
 */
static int GenStrataHashVal (struct STRATUM_ELT* aLayer)
{
    unsigned int        res;

    res = (unsigned int) aLayer->stratumLen;
    res += aLayer->dataBlob[0].Blob.pathID << 3;
    res += aLayer->dataBlob[0].Blob.tripCount;
    res += aLayer->dataBlob[1].Blob.pathID << 1;
    res += aLayer->dataBlob[1].Blob.tripCount;
    res &= (SLHSIZE - 1);       // 0x3ffff

    return (int) res;
}
/*
 * ProcessStratum - similar to ProcessPath - there are cycles of strata.
 */
#define STRATA_THRESHOLD_LENGTH 4
static void ProcessStratum (struct STRATUM_ELT* aLayer)
{
    int hIdx;
    int i;
    struct STRATALAYER_ENT*     walker;
    struct STRATALAYER_ENT*     newH;
    uint64_t       lTrips;
    int         strataIdent;

    strataIdent = 0;

        /*
         * Storing all of the strata and trying to identify them is difficult, as
         * the number of combinations is huge (9e6 for one ref run of one of the SPEC
         * CPU 2006 benchmarks)  However...
         * It appears that the number of possible hits for lengths 2,3,4 may be a huge win
         * Now to measure it.  Statistically, if the overhead of length 3,4 is minimal compared
         * to that of length==2, then it worth trying.
         * Oct 2, 2007 - revert back to processing all; hand off final write and
         * strata/count compression work to a separate server process
         * Nov 2, 2007 - subsequent runs demonstrate that the performance and memory
         * consumed by doing so is still unacceptable.
         * So try tracking all with length <= 8
         * Nov 4: 8 is too slow on bzip2/train.  More than 12 hours.  Try a setting of 4
         */

    lTrips = aLayer->tripCount;
    if (aLayer->stratumLen > STRATA_THRESHOLD_LENGTH) {
        // If it is too long, give it an ID of zero, and simply write it out
        goto done;
    }

    hIdx = GenStrataHashVal (aLayer);

    walker = SLHash[hIdx];

    while (walker != NULL) {
        if ((walker->item->addHash == aLayer->addHash)
                && (walker->item->stratumLen == aLayer->stratumLen)
                && (walker->item->dataBlob[0].l_Blob == aLayer->dataBlob[0].l_Blob)
                && (walker->item->dataBlob[1].l_Blob == aLayer->dataBlob[1].l_Blob)
                ) {
                // Need to check all remaining dataBlob items
            for (i = 2; i < aLayer->stratumLen; i++) {
                if (walker->item->dataBlob[i].l_Blob != aLayer->dataBlob[i].l_Blob) {
                    break;
                }
            }

                // If a complete match, bump count on existing, discard new.
            if (i == aLayer->stratumLen) {
                strataIdent = walker->item->stratumID;
                walker->item->tripCount += lTrips;
                goto done;
            }
        }       // Partial match
        walker = walker->next;
    }   // walking the chain

        // Got here, means no pre-existing item.  Push onto the FRONT(!) of the
        // table assuming temporal locality will help with performance.
        // Also means we don't have to track a 'prev' pointer.
    if (walker == NULL) {
        struct STRATUM_ELT*  newLayer;
        int     j;

        newLayer = AllocateElt (aLayer->stratumLen);
        newLayer->stratumLen = aLayer->stratumLen;
        newLayer->addHash = aLayer->addHash;
        newLayer->tripCount = aLayer->tripCount;
        for (j = 0; j < aLayer->stratumLen; j++) {
            newLayer->dataBlob[j] = aLayer->dataBlob[j];
        }
        stratumStats.uniquelayer0cycles++;
        strataIdent = stratumStats.uniquelayer0cycles;
        newLayer->stratumID = stratumStats.uniquelayer0cycles;
        newH = (struct STRATALAYER_ENT*) malloc (sizeof (*newH));
        newH->item = newLayer;
        newH->next = SLHash[hIdx];
        SLHash[hIdx] = newH;
    }

done: ;
    WriteStrataToCompressionServer (strataIdent, lTrips, WRITE_STRATA);
defer_write: ;
    if (finalFlush == 1) {
        WriteStrataToCompressionServer (0, 0, FLUSH_STRATA);
    }

    return;
}

static void FillStrataElt (struct STRATUM_ELT* theElt, unsigned short sLen, unsigned short addH)
{
    int j;

        // see notes in pathprof.h - minimum length is 2
    if (sLen < 2) {
        fprintf (s_LogF, "FillStrataElt error: length == %ld - using length 2\n", sLen);
        sLen = 2;
    }
    theElt->stratumID = 0;
    theElt->stratumLen = sLen;
    theElt->addHash = addH;
    theElt->tripCount = 1;
    for (j = 0; j < sLen; j++) {
        theElt->dataBlob[j] = s_tmpStratum[j];
    }
}

/*
 * Assume this is effectively ExtendPath but for paths rather than Bb sequences
 * Incoming name does not have the relevant suffixes attached for the strata db, and
 * raw strata dump info
 */
void ExtendStrataLayer (union SPB_U u_Pair)
{
    int         i;
    struct STRATUM_ELT* newSE;

    for (i = 0; i < VLPATHLIM; i++) {
        if (s_tmpStratum[i].Blob.pathID == 0x0) {
                // Not on the strata layer yet, extend it
            s_tmpStratum[i].l_Blob = u_Pair.l_Blob;
            s_tmpStratum[i + 1].Blob.pathID = 0x0;
            s_curStrataLen = i + 1;
            s_addHash += u_Pair.Blob.pathID;
            goto done;
        } else if (s_tmpStratum[i].l_Blob == u_Pair.l_Blob) {
                // Completed a cycle.  Adjust stats, hand off for further processing
            stratumStats.layer0cycles++;
            if (s_curStrataLen > stratumStats.longestLayer) {
                stratumStats.longestLayer = s_curStrataLen;
            }
            stratumStats.strataDistrib[s_curStrataLen]++;

                // Very first time only.  Allocate a maximum sized entry for cacheing.
            if (s_prevStratum == NULL) {
                s_prevStratum = AllocateElt (VLPATHLIM);
                FillStrataElt (s_prevStratum, s_curStrataLen, s_addHash);
                goto reset;
            }

                // Check for immediate match
           if ((s_prevStratum->addHash == s_addHash) && (s_prevStratum->stratumLen == s_curStrataLen)) {
                int j;

                for (j = 0; j < s_curStrataLen; j++) {
                    if (s_prevStratum->dataBlob[j].l_Blob != s_tmpStratum[j].l_Blob) {
                        ProcessStratum (s_prevStratum);
                        FillStrataElt (s_prevStratum, s_curStrataLen, s_addHash);
                        goto reset;
                    }
                }
                // Complete match if we got here.  Simply bump trip count
                s_prevStratum->tripCount++;
                stratumStats.consecutive++;
                stratumStats.strataMatches[s_prevStratum->stratumLen]++;
                goto reset;
            }

                // If we got here, not a match
            ProcessStratum (s_prevStratum);
            FillStrataElt (s_prevStratum, s_curStrataLen, s_addHash);
            goto reset;

reset: ;
            s_tmpStratum[0].l_Blob = u_Pair.l_Blob;
            s_tmpStratum[1].Blob.pathID = 0x0;
            s_addHash = u_Pair.Blob.pathID;
            s_curStrataLen = 1;
            goto done;
        }       // We had a match, meaning a new stratum layer
    }   // Walk up the the limit of the s_tmpStratum

        /*
         * Not found, and we are at the temp table end.  A very long strata layer.
         * Reset as if we are starting a new path.  Note that this smears the
         * statistics because we are now starting to look at a shorter layer
         * NOTE: if this happens a lot, it means that there is a really obnoxious case
         * and VPATHLIM probably needs to be doubled.
         */

    if (i == VLPATHLIM) {
        fprintf (s_LogF, "ExtendStrataLayer - cycle is too long (%d) - will reset to length 2\n", VLPATHLIM);
        stratumStats.VLLayers++;
        s_tmpStratum[0].l_Blob = s_tmpStratum[i - 2].l_Blob;
        s_tmpStratum[1].l_Blob = s_tmpStratum[i - 1].l_Blob;

        s_tmpStratum[2].Blob.pathID = 0x0;
        s_curStrataLen = 2;
        s_addHash = s_tmpStratum[0].Blob.pathID + s_tmpStratum[1].Blob.pathID;
        goto done;
    }

        // Consistency error.  We never should have gotten here.
    fprintf (s_LogF, "ExtendStrataLayer error - very bad problem: pathIdent == %ld, i == %ld\n", u_Pair.Blob.pathID, i);

done: ;
    return;
}

void StrataInit (char *baseName)
{
    struct STRATUM_ELT  checkme;
    int         rcode;
    char        idxFName[PATH_MAX];
    char        datFName[PATH_MAX];
    char        logFName[PATH_MAX];
    int shmKey;
    int memSize;
    char        strataInfoDBFile[PATH_MAX];
    char        rawStrataFile[PATH_MAX];

    snprintf (logFName, sizeof(logFName) - 1, "%s.stratalog.txt", baseName);
    s_LogF = fopen (logFName, "w+");
    if (s_LogF == NULL) {
        fprintf (stderr, "%s: Unable to open logfile [%s] --> errno: %ld\n",
                g_argV0, logFName, errno);
    }

    if (sizeof (s_addHash) != sizeof (checkme.addHash)) {
        /*
         * If they do not match, the strata layers will not be calculated correctly,
         * as unique, resulting in a lot more layers (and consequently db writes)
         * which will slow the collection down significantly.
         */
        fprintf (s_LogF, "Serious problem - local hash size in strata.c is mismatched - aborting\n");
        exit (1);
    }

    finalFlush = 0;
    memset ((void*) &s_tmpStratum, 0, sizeof (s_tmpStratum));
    memset ((void*) &s_addHash, 0, sizeof (s_addHash));
    memset ((void*) &stratumStats, 0, sizeof (stratumStats));
    s_curStrataLen = 0;
    s_prevStratum = NULL;
    SLHash = (struct STRATALAYER_ENT**) calloc (SLHSIZE, sizeof (*SLHash));

        // strata PureDB for strata info
    snprintf (idxFName, sizeof(idxFName) - 1, "%s%s.s_tmpidx", baseName, STRATA_INFO_DB_NAME);
    snprintf (datFName, sizeof(datFName) - 1, "%s%s.s_tmpdat", baseName, STRATA_INFO_DB_NAME);
    snprintf (strataInfoDBFile, sizeof(strataInfoDBFile) - 1, "%s%s.s_dbp", baseName, STRATA_INFO_DB_NAME);

    rcode = puredbw_open (&strataDB_pureW, idxFName, datFName, strataInfoDBFile);
    if (rcode != 0) {
        fprintf (s_LogF, "Bad open of strata puredb: %d - aborting\n", rcode);
        exit (1);
    }

        // raw resultant data file
    snprintf (rawStrataFile, sizeof(rawStrataFile) - 1, "%s%s", baseName, RAW_STRATA_NAME);

    memset ((void*) &s_strataBuf, 0, sizeof (s_strataBuf));
    strcpy (s_strataBuf.dataHdr.dataFileBaseName, rawStrataFile);
    s_rawIdx = 0;

}

static void CheckHashFn()
{
    struct STRATALAYER_ENT*  walker;
    int badHashfn;
    int i;

    badHashfn = 0;
    for (i = 0; i < SLHSIZE; i++) {
        int     sum;

        walker = SLHash[i];
        sum = 0;
        while (walker) {
            sum++;
            walker = walker->next;
        }

        // If we see a depth of more than about 5% that means we have a problem
        if ((sum > 10) && (sum * 20 > stratumStats.uniquelayer0cycles)) {
            badHashfn = 1;
            break;
        }
    }   // Each slot

    if (badHashfn != 0) {
        fprintf (s_LogF, "*****************************\n");
        fprintf (s_LogF, "Poor Strata Hash table distribution - consider reworking GenHashVal()\n"
                "Important info: (abnormally deep slots)\nSlot\t(Hex)\tDepth\n");
        for (i = 0; i < SLHSIZE; i++) {
            int     sum;

            walker = SLHash[i];
            sum = 0;
            while (walker) {
                sum++;
                walker = walker->next;
            }
                // Report anything deeper than 0.1%
            if ((sum > 10) && ((sum * 1000) > stratumStats.uniquelayer0cycles)) {
                fprintf (s_LogF, "%d\t%x\t%d\n", i, i, sum);
            }
        }
        fprintf (s_LogF, "*****************************\n");
    } else {
        fprintf (s_LogF, "Info: Good Strata hash distribution: (hash function works well)\n");
    }
}

static void WriteLongToStrataDB (char *keystr, uint64_t val)
{
    const char* func = "WriteIntToStrataDB";
    uint64_t        scratchValue;
    char        dbkey[2048];
    datum       Key;
    datum       Value;
    int         rc;

    snprintf (dbkey, sizeof(dbkey), "%s", keystr);
    scratchValue = val;
    Key.dptr = dbkey;
    Key.dsize = strlen (dbkey);
    Value.dptr = (char *) &scratchValue;
    Value.dsize = sizeof (scratchValue);
    rc = puredbw_add (&strataDB_pureW, Key.dptr, Key.dsize, Value.dptr, Value.dsize);
    if (rc != 0) {
        fprintf (s_LogF, "Bad puredbw_add in %s for key: [%s]: %d\n", func, keystr, rc);
    }
}

static void WriteStructToStrataDB (char *keystr, char *valBuf, uint64_t valBufSize)
{
    const char* func = "WriteStructToStrataDB";
    uint64_t        scratchValue;
    char        dbkey[2048];
    datum       Key;
    datum       Value;
    int         rc;

    snprintf (dbkey, sizeof(dbkey), "%s", keystr);
    Key.dptr = dbkey;
    Key.dsize = strlen (dbkey);
    Value.dptr = (char *) valBuf;
    Value.dsize = valBufSize;
    rc = puredbw_add (&strataDB_pureW, Key.dptr, Key.dsize, Value.dptr, Value.dsize);
    if (rc != 0) {
        fprintf (s_LogF, "Bad puredbw_add in %s for key: [%s]: %d\n", func, keystr, rc);
    }
}

/*
 * Unlike the pathsdb, we really don't care much about which ones are the hottest
 * So we'll operate directly on the hash table instead of a sorted array...
 */
static void WriteStrataDB()
{
    int i;
    int count;
    struct STRATALAYER_ENT*  walker;
    datum   Key;
    datum   Value;
    char        dbkey[2048];

    count = stratumStats.uniquelayer0cycles;    // A sanity checker

    WriteLongToStrataDB (KEY_VERNO, KEY_VERSION);     // for expandability
    WriteLongToStrataDB (KEY_NUMPATHS, stratumStats.uniquelayer0cycles);

        /*
         * The finished stratumStats struct
         */
    WriteStructToStrataDB (KEY_STRATUMSTATS, (char*) &stratumStats, sizeof(stratumStats));

        /*
         * Write the individual stratum structures
         */
    for (i = 0; i < SLHSIZE; i++) {
        walker = SLHash[i];
        while (walker) {
            struct STRATUM_ELT* curS;
            int                 strataSize;
            

            curS = walker->item;
                // Remember that the first two are free, and the minimum len is two.
            strataSize = sizeof(struct STRATUM_ELT)
                        + (sizeof(struct SMALL_PATH_BLOB) * (curS->stratumLen - 2));
            snprintf (dbkey, sizeof(dbkey), "%s:%d", KEY_STRATUMID, curS->stratumID);
            WriteStructToStrataDB (dbkey, (char*) curS, strataSize);
            walker = walker->next;
            count--;
        }       // All entries per slot
    }   // Each hash table slot

    if (count != 0) {
        fprintf (s_LogF, "Consistency error: count is %d, should be zero\n");
    }
}

static void CloseStrataDB()
{
    (void) puredbw_close (&strataDB_pureW);
    puredbw_free (&strataDB_pureW);
}

void StrataFini()
{
    int i;
    struct STRATUM_ELT* newSE;
    uint64_t        dbival;
    int         rc = 0;
    int         checksum = 0;

    ProcessStratum (s_prevStratum);
    FillStrataElt (s_prevStratum, s_curStrataLen, (unsigned short) s_addHash);
    finalFlush = 1;
    ProcessStratum (s_prevStratum);

    CheckHashFn();

    if (s_curStrataLen > stratumStats.longestLayer) {
        stratumStats.longestLayer = s_curStrataLen;
    }
    stratumStats.strataDistrib[s_curStrataLen]++;

    WriteStrataDB();
    CloseStrataDB();

        // And free up all dynamically allocated memory
    for (i = 0; i < SLHSIZE; i++) {
        struct STRATALAYER_ENT* walker;
        struct STRATALAYER_ENT* next;

        walker=SLHash[i];
        while (walker) {
            next = walker->next;
            free (walker);
            walker = next;
            checksum++;
        }
    }
    free (SLHash);
    free (s_prevStratum);

    if (checksum != stratumStats.uniquelayer0cycles) {
        fprintf (s_LogF, "Error: freed incorrect number of strata entries\n");
    }

        // Dump strata stats
    fprintf (s_LogF, "---------------- Strata Info ----------------\n"
                "Approximate size of total (uncompressed) data written: %ld in %ld dbm_stores\n"
                "Approximate size of total (compressed) data written: %ld\n"
                "Cycles: %ld\tUnique Strata: %ld\tLongest Chain: %ld\tConsecutive hits: %ld\n",
                stratumStats.totalSize, stratumStats.numDBEntries,
                stratumStats.totalCompressedSize,
                stratumStats.layer0cycles, stratumStats.uniquelayer0cycles,
                stratumStats.longestLayer, stratumStats.consecutive);
    fprintf (s_LogF, "Length\tCount\tConsecutive hits\n");
    for (i = 0; i < VLPATHLIM; i++) {
        if (stratumStats.strataDistrib[i] != 0) {
            fprintf (s_LogF, "%ld\t%ld\t%ld\n", i, stratumStats.strataDistrib[i],
                        stratumStats.strataMatches[i]);
        }
    }
    fclose (s_LogF);
}
