/*
 *      Copyright Richard Gorton, 2007-2011 All rights reserved
 *
 *      Permission is granted to use, duplicate, and
 *      publish this text and program as long as it
 *      includes this entire comment block and limited
 *      rights reference.
 */

/*
 * strataserver.c - A DWPPP (Dynamic Whole Program Path Profiler) server to analyze
 *      strata (cycles of paths).  This takes advantage of processor dies having at
 *      least 2 cores.  The incoming data is comprised of path ids.
 *      The server will analyze and algorithmically reduce the incoming content and
 *      write out the result to another server - the compression server.
 *
 *      Written to compensate for suck-oh performance when all processing is
 *      done inside a PIN tool (and because multi-threaded PIN tools don't seem
 *      to work particularly well).
 *
 * Assumptions:
 *      Only one sender of data at a time
 *      Uses sysV shmem as the data transfer method
 *      This server MUST be started AFTER the compression server.
 */

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <zlib.h>

#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>

#include "pathprof.h"
#include "servercomm.h"

/*
 * Organization of the resultant data:
 *      a strata_db database (pureDB format) of each of the strata
 */
/*
 * server options: none
 */

static struct DF_INFO   dataPair;
static struct SHARED_MEM*       s_sharedMem = NULL;

struct SHARED_MEM*      g_compressSharedMem = NULL;
char*           g_argV0;
uint64_t   g_insnEpoch;
uint64_t   g_pathEpoch;

static void cleanup()
{
}

static void cleanup_exit()
{
    fprintf (stderr, "Aborting\n");
    cleanup();
    exit (1);
}

static void processArgs (int argc, char *argv[])
{
}

/*
 * Analyze and algorithmically compress the strata (cyclical path sequences)
 * Generate a strata info database file containing the details of each strata.
 */
void processPayload (struct SHARED_MEM* memData)
{
    union SPB_U*     theBlobs;
    uint64_t        i;
    uint64_t        numRecs;

    numRecs = memData->dataHdr.payloadSize / sizeof(struct SMALL_PATH_BLOB);
// fprintf (stderr, "processPayload of %ld records\n", numRecs);
    g_insnEpoch = memData->dataHdr.startInsnEpoch;
    g_pathEpoch = memData->dataHdr.startPathEpoch;
    theBlobs = (union SPB_U *) &memData->rawDBBuf[0];
    for (i = 0; i < numRecs; i++) {
        ExtendStrataLayer (theBlobs[i]);
    }
}

int main(int argc, char *argv[])
{
    ssize_t     rdata;
    int         clientlen;
    int         rc;
    int         exitcleanly = 0;
    char        idxFName [PATH_MAX];
    char        dataFName [PATH_MAX];
    struct SHARED_MEM*       s_sharedMem = NULL;

    g_argV0 = argv[0];
    processArgs (argc, argv);
    signal (SIGINT, cleanup_exit);

    unlink (STRATA_MMAP_FILE);
    s_sharedMem = OpenMMapFile (argv[0], SERVER_OPEN_MMAPFILE, STRATA_MMAP_FILE);
    if (s_sharedMem == NULL) {
        cleanup_exit();
    }

    g_compressSharedMem = OpenMMapFile (g_argV0, CLIENT_OPEN_MMAPFILE, COMPRESSOR_MMAP_FILE);
    if (g_compressSharedMem == NULL) {
        fprintf (stderr, "%s: Unable to open %s as client\n", g_argV0, COMPRESSOR_MMAP_FILE);
        cleanup_exit();
    }

    dataPair.state = STATE_NONE;
        // And mark region as ready for use
    MarkReadyToFillData(s_sharedMem);

    while (!exitcleanly) {
        struct SHARED_MEM       localSHM;

        WaitUntilDataFilled(s_sharedMem);

        // Data in the shared memory buffer, ready to process

        // Only relevant once deployed in the field (if at all) and the format changes
        if (s_sharedMem->dataHdr.version != THIS_VERSION) {
            fprintf (stderr, "%s: received data from an unsupported version client: %d\n",
                        argv[0], s_sharedMem->dataHdr.version);
            continue; 
        }

        /*
         * At this point, start trying to be as concurrent as possible.  Make a local
         * copy of the content, and operate on that, and then mark the shared memory
         * region as available to write new data into.  Assume that it will be almost
         * completely full, so simply copy the entire thing.  The only time it will be
         * sparse is on the final set of data.
         * Once the copy is made, be sure to _not_ reference the shared memory region
         */
        memcpy ((void*) &localSHM, (void*) s_sharedMem, sizeof (localSHM));
        MarkReadyToFillData(s_sharedMem);

        // Open operations
        // Very simple for strata analyzer: initialize strata tables.
        if (localSHM.dataHdr.operation & (OP_OPEN_APPEND|OP_OPEN_TRUNC)) {
            if (dataPair.state == STATE_OPENED) {
                if (strcmp (localSHM.dataHdr.dataFileBaseName, dataPair.baseFileName) == 0) {
                    fprintf (stderr, "%s: [%s] is already open... ignoring open op\n",
                                argv[0], localSHM.dataHdr.dataFileBaseName);
                    goto skip_open;
                } else {
                    fprintf (stderr,
                                "%s: Unable to open (append) to [%s] - another file pair is in use: [%s]\n",
                                argv[0], localSHM.dataHdr.dataFileBaseName, dataPair.baseFileName);
                }
                continue;
            }
            StrataInit (localSHM.dataHdr.dataFileBaseName);
            dataPair.state = STATE_OPENED;
        }

skip_open: ;
        // Process the data payload (if any)
        processPayload (&localSHM);

        if (localSHM.dataHdr.operation & (OP_FLUSH_CLOSE | OP_DELETE)) {
            StrataFini();
            dataPair.state = STATE_CLOSED;
        }

        // Delete not meaningful for strata_db file
        if (localSHM.dataHdr.operation & OP_DELETE) {
            dataPair.indexFileName[0] = 0x0;
            dataPair.dataFileName[0] = 0x0;
            dataPair.state = STATE_NONE;
        }
    }

done: ;

    cleanup();
    exit (0);
}
